import React from 'react';
import PropTypes from 'prop-types';
import classNames from 'classnames/bind';
import AnueLink from '../../utils/anueLink';

import styles from './ScrollingQuote.scss';

const cx = classNames.bind(styles);

class ScrollingQuote extends React.Component {
  static propTypes = {
    inverted: PropTypes.bool,
    speed: PropTypes.number,
    quotes: PropTypes.arrayOf(
      PropTypes.shape({
        symbol: PropTypes.string.isRequired,
        name: PropTypes.oneOfType([PropTypes.string, PropTypes.element]).isRequired,
        value: PropTypes.string.isRequired,
        netChange: PropTypes.string.isRequired,
        changeInPercentage: PropTypes.string.isRequired,
      })
    ),
    handleScrollEnd: PropTypes.func.isRequired,
    env: PropTypes.string.isRequired,
    refresh: PropTypes.number,
    bordered: PropTypes.bool,
  };

  static defaultProps = {
    inverted: false,
    speed: 1,
    quotes: [],
    refresh: 0,
    bordered: false,
  };

  componentDidMount() {
    this.scrollerElementWidth = this.getScrollerElementWidth();
    this.shift = this.getWrapperElementWidth();

    this.startScroll();
  }

  componentDidUpdate(prevProps) {
    if (this.props.quotes !== prevProps.quotes || this.props.refresh !== prevProps.refresh) {
      this.scrollerElementWidth = this.getScrollerElementWidth();

      this.startScroll();
    }
  }

  componentWillUnmount() {
    this.stopScroll();
  }

  getScrollerElementWidth = () => this.scrollerElement.getBoundingClientRect().width;

  getWrapperElementWidth = () => this.scrollerElement.parentNode.getBoundingClientRect().width;

  setScrollerElement = element => {
    this.scrollerElement = element;
  };

  scroll = () => {
    const { handleScrollEnd, speed } = this.props;

    if (this.shift < -this.scrollerElementWidth) {
      this.shift = this.getWrapperElementWidth();
      this.scrollerElement.style.transform = `translateX(${this.shift}px)`;

      handleScrollEnd();

      return;
    }

    this.shift = this.shift - speed;
    this.scrollerElement.style.transform = `translateX(${this.shift}px)`;

    if (window && window.requestAnimationFrame) {
      this.rafId = window.requestAnimationFrame(this.scroll);
    }
  };

  startScroll = () => {
    this.stopScroll();

    if (window && window.requestAnimationFrame) {
      this.rafId = window.requestAnimationFrame(this.scroll);
    }
  };

  stopScroll = () => {
    if (window && window.cancelAnimationFrame && this.rafId) {
      window.cancelAnimationFrame(this.rafId);
    }
  };

  render() {
    const { inverted, quotes, env, bordered } = this.props;
    const anueUtilLink = new AnueLink(env);

    return (
      <div
        className={cx('wrapper', { inverted, 'wrapper--border': bordered })}
        onMouseEnter={this.stopScroll}
        onMouseLeave={this.startScroll}
      >
        <div className={cx('scroller')} ref={this.setScrollerElement}>
          {quotes.map(({ symbol, name, value, netChange, changeInPercentage }) => {
            const sign = netChange[0];
            const positive = sign === '+';
            const negative = sign === '-';
            const link = anueUtilLink.generateStockLinkBySymbol(symbol);

            return (
              <span key={symbol} className={cx('quote', { positive, negative })}>
                <span className={cx('name')}>
                  {link ? <a href={anueUtilLink.generateStockLinkBySymbol(symbol)}>{name}</a> : name}
                </span>
                <span>{value}</span>
                <span className={cx('mark')} />
                <span className={cx('change')}>{netChange}</span>
                <span className={cx('change')}>{changeInPercentage}</span>
              </span>
            );
          })}
        </div>
      </div>
    );
  }
}

export default ScrollingQuote;
